/* ISA card device driver, common interface.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)
 
   v 0.11.0  3  Sep 2013 by Winson Chen
     Gives support to linux kernel 3.X.X

   v 0.10.0  10 Aug 2011 by Golden Wang
     Gives support to linux kernel 2.6.x.

   v 0.9.0  2 May 2003 by Reed Lai
     Gives support to A-8111.

   v 0.8.0 29 Apr 2003 by Reed Lai
     Gives support to A-826.

   v 0.7.0  9 Apr 2003 by Reed Lai
     Gives support to A-822.

   v 0.6.0 18 Nov 2002 by Reed Lai
     Gives support to P8R8DIO and P16R16DIO.
     Uses slab.h in place of malloc.h which has been deprecated.
	 Complies to the kernel module license check.
     Gives support to kernel 2.4.

   v 0.5.0 27 Feb 2002 by Reed Lai
     Use a new flag type for device information (ixisa_devinfo.)

   v 0.4.0  4 Feb 2002 by Reed Lai
     Add A823.

   v 0.3.0 21 Jan 2002 by Reed Lai
     Add DIO-144.

   v 0.2.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.

   v 0.1.0 2001.10.24 by Reed Lai
     Modified structure proc_dir_entry ixisa_procfile for Linux kernel 2.4
     compatibility.

   v 0.0.0 2001.02.13 by Reed Lai
     create, blah blah... */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */
#include "ixisa.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        #include <linux/devfs_fs_kernel.h>
        #include <linux/init.h>
#endif

/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
        #if CONFIG_MODVERSIONS==1
        #define MODVERSIONS
        #include <linux/modversions.h>
        #endif
/* Additional */
// #include <linux/fs.h> /* file system here */
        #include <linux/wrapper.h>
#endif

	/* for compatibility with future version of Linux */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>		/* function put_user() */
#endif

/* support /proc file */
#include <linux/proc_fs.h>

/* using I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
#include <linux/slab.h>

/* Local */

#define MODULE_NAME "ixisa"

#ifdef MODULE_LICENSE
MODULE_AUTHOR("Golden Wang <service@icpdas.com>");
MODULE_DESCRIPTION("ICPDAS ISA card drivers, common interface");
MODULE_LICENSE(ICPDAS_LICENSE);
#endif							/* MODULE_LICENSE */

/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
EXPORT_SYMBOL_GPL(ixisa_dev);
EXPORT_SYMBOL_GPL(ixisa_copy_devinfo);
#else
EXPORT_SYMBOL_NOVERS(ixisa_dev);
//EXPORT_SYMBOL_NOVERS(ixisa_major);
EXPORT_SYMBOL_NOVERS(ixisa_copy_devinfo);
#endif

/* functions are declared here first */
static int ixisa_open(struct inode *, struct file *);
int ixisa_release(struct inode *, struct file *);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long ixisa_ioctl(struct file *, unsigned int , unsigned long );
#else
int ixisa_ioctl(struct inode *, struct file *, unsigned int , unsigned long );
#endif

/* device file operaitons information */
static struct file_operations fops = {
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        owner:THIS_MODULE,
#endif
        open:ixisa_open,
        release:ixisa_release,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
        unlocked_ioctl:ixisa_ioctl
#else
        ioctl:ixisa_ioctl
#endif
};

int minor = 0;  //addmyself

/* Card definitions */
struct ixisa_carddef ixisa_card[] = {
/*  id          present  module     name         */
	{A626, 0, "ixa626", A626_NAME},
	{A628, 0, "ixa628", A628_NAME},
	{A8111, 0, "ixa8111", A8111_NAME},
	{A812PG, 0, "ixa812pg", A812PG_NAME},
	{A821, 0, "ixa821", A821_NAME},
	{A822, 0, "ixa822", A822_NAME},
	{A823, 0, "ixa823", A823_NAME},
	{A826, 0, "ixa826", A826_NAME},
	{DIO24, 0, "ixdio24", DIO24_NAME},
	{DIO48, 0, "ixdio48", DIO48_NAME},
	{DIO64, 0, "ixdio64", DIO64_NAME},
	{DIO96, 0, "ixdio96", DIO96_NAME},
	{DIO144, 0, "ixdio144", DIO144_NAME},
	{ISO730, 0, "ixiso730", ISO730_NAME},
	{ISO813, 0, "ixiso813", ISO813_NAME},
	{ISOAD32, 0, "ixisoad32", ISOAD32_NAME},
	{ISOC64, 0, "ixisoc64", ISOC64_NAME},
	{ISODA16, 0, "ixisoda16", ISODA16_NAME},
	{ISOLD, 0, "ixisold", ISOLD_NAME},
	{ISOP32C32, 0, "ixisop32c32", ISOP32C32_NAME},
	{ISOP64, 0, "ixisop64", ISOP64_NAME},
	{P8R8DIO, 0, "ixp8r8dio", P8R8DIO_NAME},
	{P16R16DIO, 0, "ixp16r16dio", P16R16DIO_NAME},
	{ENCODER300, 0, "ixenc300", ENCODER300_NAME},
	{SERVO300, 0, "ixservo300", SERVO300_NAME},
	{STEP200, 0, "ixstep200", STEP200_NAME},
	{TMC10, 0, "ixtmc10", TMC10_NAME},
	{0, 0, "", "UNKNOW"},
};

int _card_id(char *name, int new)
{
	/* get Card's id by name
	 *
	 * Arguments:
	 *   name     string of card name
	 *   new      integer to flag new card
	 *
	 * Returned:  card id, 0 not found */

	int i = 0;
	while (ixisa_card[i].id) {
		if (!strcmp(name, ixisa_card[i].name)) {
			if (new)
			{
				++(ixisa_card[i].present);
				
			}
			/* yeh, one more, check in */
			return ixisa_card[i].id;
		}
		++i;
	}
	return 0;
}

void *_cardname(int id, int add_present)
{
	/* (export)
	 *
	 * get card name by sub-ids
	 *
	 * Arguments:
	 *   id       composed sub-ids, see ixpio.h
	 *
	 * Returned:  pointer to the name string that coresponds the given sub
	 *            id, if there is no card name found, returns 0 */

	int i = 0;

	while (ixisa_card[i].id) {
		if (ixisa_card[i].id == id) {
			if (add_present)
				++(ixisa_card[i].present);
			return ixisa_card[i].name;
		}
		++i;
	}
	return 0;
}

__inline__ void *_isa_cardname(int id)
{
	return _cardname(id, 0);
}

//int ixisa_major = 0;
int ixisa_major = DEVICE_MAJOR;
int ixisa_minor = DEVICE_MINOR;
	/* Global major number */
ixisa_devinfo_t *ixisa_dev;
	/* pointer to the cards' list */

/*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int __devinit ixisa_board_init (ixisa_devinfo_t *dev, int index)
{
	int err, devno;
	devno = MKDEV(ixisa_major, ixisa_minor + index);
	dev->cdev = kmalloc(sizeof(struct cdev), GFP_KERNEL);
	memset(dev->cdev, 0, sizeof(struct cdev));
	cdev_init(dev->cdev, &fops);

	dev->cdev->owner = THIS_MODULE;
	dev->cdev->ops = &fops;

	err = cdev_add(dev->cdev, devno, 1);

	if (err)
	{
		printk("Error %d adding ixisa device\n", err);
		return FAILURE;
	}

	return SUCCESS;
}
#endif
*/

void _del_dev(void)
{
	/* release memory of card list
	 *
	 * Arguments: none
	 *
	 * Returned:  void */

	ixisa_devinfo_t *dev, *prev;

	dev = ixisa_dev;
	while (dev) {
		prev = dev;
		dev = dev->next;
		kfree(prev);
	}
}

void ixisa_copy_devinfo(ixisa_devinfo_t * dst, ixisa_devinfo_t * src)
{
	dst->id = src->id;
	dst->no = src->no;
	dst->irq = src->irq;
	dst->dma.nr = src->dma.nr;
	dst->base = src->base;
	strncpy(dst->name, src->name, CNL);
}

int _add_dev(struct ixisa_devinfo *new_dev)
{
	/* add card to list
	 *
	 * Arguments: pointer to structure ixisa_devinfo of new device
	 *
	 * Returned:  SUCCESS or FAILURE */

	ixisa_devinfo_t *dev, *prev, *prev_f;
	new_dev->id = _card_id(new_dev->name, 1);

	if (!new_dev->id)
		return FAILURE;

	if (ixisa_dev) {
		/* ixisa device list is already followed */
		if (ixisa_dev->id == new_dev->id)
			prev_f = ixisa_dev;
		else
			prev_f = 0;
		prev = ixisa_dev;
		dev = prev->next;
		while (dev) {
			/* seek the tail of list */
			if (dev->id == new_dev->id)
				prev_f = dev;
			/* last previous device in family */
			prev = dev;
			dev = dev->next;
		}
		dev = prev->next = kmalloc(sizeof(ixisa_devinfo_t), GFP_KERNEL);
		memset(dev, 0, sizeof(ixisa_devinfo_t));
		dev->prev = prev;
		/* member the previous address */

		if (prev_f) {
			dev->prev_f = prev_f;
			/* member the previous family device */
			prev_f->next_f = dev;
			/* member the next family device */
		}
	} else {
		/* pio device list is empty, initiate */
		ixisa_dev = kmalloc(sizeof(ixisa_devinfo_t), GFP_KERNEL);
		memset(ixisa_dev, 0, sizeof(ixisa_devinfo_t));
		dev = ixisa_dev;
	}
	ixisa_copy_devinfo(dev, new_dev);
	return SUCCESS;
}

ixisa_devinfo_t *_align_to_minor(int minor)
{
	ixisa_devinfo_t *dp;
	for (dp = ixisa_dev; dp && dp->no != minor; dp = dp->next);
	return dp;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long ixisa_ioctl(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
#else
int ixisa_ioctl(struct inode *inode, struct file *file,	unsigned int ioctl_num, unsigned long ioctl_param)
#endif
{
	/* This function is called whenever a process tries to do and IO
	 * control on IXISA device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	switch (ioctl_num) {
	case IXISA_ADD_DEV:
		/* add card to list */
		return (_add_dev((ixisa_devinfo_t *) ioctl_param));
	case IXISA_DEL_DEV:
		/* delete card from list */
		return -EINVAL;
	default:
		/* pass request to service module */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
		dp = _align_to_minor(MINOR(minor));

		if (!dp) {
                        /* if not found, get out */
                        return -EINVAL;
                }
                if (!dp->fops || !dp->fops->unlocked_ioctl) {
                        return -EINVAL;
                }
                return (dp->fops->unlocked_ioctl) (file, ioctl_num, ioctl_param);

#else
		dp = _align_to_minor(MINOR(inode->i_rdev));

		if (!dp) {
			/* if not found, get out */
			return -EINVAL;
		}
		if (!dp->fops || !dp->fops->ioctl) {
			return -EINVAL;
		}
		return (dp->fops->ioctl) (inode, file, ioctl_num, ioctl_param);
#endif
	}
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
int ixisa_release(struct inode *inode, struct file *file)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int ixisa_release(struct inode *inode, struct file *file)
#else
static void ixisa_release(struct inode *inode, struct file *file)
#endif
{
	/* This function is called whenever a process attempts to closes the
	 * device file. It doesn't have a return value in kernel version 2.0.x
	 * because it can't fail (you must always be able to close a device).
	 * In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;
	int minor;

	minor = MINOR(inode->i_rdev);

	if (!minor)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		return SUCCESS;
	/* minor == 0, dummy, do nothing */
#else
		return;
#endif

	dp = _align_to_minor(minor);

	if (!dp->fops || !dp->fops->release) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		return -EINVAL;
#endif
	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
		return (dp->fops->release) (inode, file);
#else
		(dp->fops->release) (inode, file);
#endif
	}
}

static int ixisa_open(struct inode *inode, struct file *file)
{
	/* This function is called whenever a process attempts to open the
	 * device file
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;
	minor = MINOR(inode->i_rdev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
	file->private_data = &minor;
#endif
	if (!minor)
		return SUCCESS;
	/* minor == 0, dummy, do nothing */

	dp = _align_to_minor(minor);

	if (!dp->fops || !dp->fops->open) {
		return -EINVAL;
	}
	return (dp->fops->open) (inode, file);
}

void ixisa_cleanup(void)
{
	/* remove proc file */
	ixisa_proc_exit();

	/* remove device file operations */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        unregister_chrdev(ixisa_major, DEVICE_NAME);
#else
	int unr;
	unr = devfs_unregister_chrdev(ixisa_major, DEVICE_NAME);
	if (unr < 0) {
		printk(KERN_INFO FAMILY
			   ": %s module_unregister_chrdev() error: %d\n", MODULE_NAME,
			   unr);
	}
#endif
	_del_dev();					/* release allocated memory */

	printk(KERN_INFO FAMILY ": %s has been removed\n", MODULE_NAME);
}

int ixisa_init(void)
{
	/* initialize this module
	 *
	 * Arguments:
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	int err;

	printk(KERN_INFO FAMILY ": %s\n", MODULE_NAME);

/*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        dev_t devt = 0;

        err = alloc_chrdev_region(&devt, DEVICE_MINOR, DEVICE_NR_DEVS, DEVICE_NAME);

        if (err < 0)
        {
		printk(KERN_WARNING "ixpio: can't get major %d\n", ixisa_major);
        }
        else
        {
		ixisa_major = MAJOR(devt);
        }
#endif
*/
	/* register device file operations  */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	ixisa_major = register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &fops);
#else
	ixisa_major = devfs_register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &fops);
#endif

	if (ixisa_major < 0) {
		printk(KERN_INFO FAMILY ": major %d failed!\n", ixisa_major);
		_del_dev();
		return ixisa_major;
	}
	printk(KERN_INFO FAMILY ": major %d.\n", ixisa_major);

	/* setup proc file */
	if ((err = ixisa_proc_init())) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
                KMSG("/proc/%s/%s %d failed!\n", ORGANIZATION, DEVICE_NAME, err);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		KMSG("%s/%s/%s %d failed!\n", proc_root.name,
			 ORGANIZATION, DEVICE_NAME, err);
#else
		KMSG("%s/%s %d failed!\n", proc_root.name, DEVICE_NAME, err);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                unregister_chrdev(ixisa_major, DEVICE_NAME);
#else
		devfs_unregister_chrdev(ixisa_major, DEVICE_NAME);
#endif
		_del_dev();
		return err;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
        KMSG("/proc/%s/%s %d.\n", ORGANIZATION, DEVICE_NAME, err);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	KMSG("%s/%s/%s %d.\n", proc_root.name, ORGANIZATION, DEVICE_NAME, err);
#else
	KMSG("%s/%s %d.\n", proc_root.name, DEVICE_NAME, err);
#endif
	return SUCCESS;
}

#if LINUX_VERSION_CODE >=KERNEL_VERSION(2,4,0)
module_init(ixisa_init);
module_exit(ixisa_cleanup);
#endif
