/* Example of interrupt handle for DIO-48.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   This example uses mode 2 that is direct trigger from CN1 Port C3.
   Set the SW1 to mode 2 as following

     +-----------------+
     |on               |
     | +-------------+ |
     | | o   o o o o | |
     | |             | |
     | |   o         | |
     | +-------------+ |
     |   1 2 3 4 5 6   |
     +-----------------+

   Connect CN1 and CN2, or CN1PC3 and CN2PC3

   v 0.2.0  15 Aug 2011 by Golden Wang
     Add "usleep(100)" to fix the bug that counting the number of interrupt error.

   v 0.1.2  2 Dec 2002 by Reed Lai
     Improves signal operating.

   v 0.1.1 18 Jan 2002 by Reed Lai
     Correct the misprinting message.

   v 0.1.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.
   
   v 0.0.0 17 Oct 2001 by Reed Lai
     create, blah blah... */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <signal.h>

#include "ixisa.h"

unsigned sig_counter;

void sig_handler(int sig)
{

	++sig_counter;
	putchar('+');

}

int main()
{
	int fd;
	char *dev_file;
	ixisa_reg_t reg;
	ixisa_signal_t sig;

	static struct sigaction act, act_old;

	sig_counter = 0;
	dev_file = "/dev/ixisa1";

	/* open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of openning device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* signal action */
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);
	if (sigaction(SIGALRM, &act, &act_old)) {
		close(fd);
		puts("Failure of signal action.");
		return FAILURE;
	}

	/* Configure CN1 ports as input */
	reg.id = IXISA_CN1CR;
	reg.value = 0x9b;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of CN1.");
		return FAILURE;
	}

	/* Configure CN2 ports as output */
	reg.id = IXISA_CN2CR;
	reg.value = 0x80;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of CN2.");
		return FAILURE;
	}

	/* signal condiction */
	sig.sid = SIGALRM;
	sig.pid = getpid();
	if (ioctl(fd, IXISA_SET_SIG, &sig)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of signal condiction.");
		return FAILURE;
	}

	/* rocking time */
	reg.id = IXISA_CN2PC;
	puts("press several space keys to IRQ, then enter to get out\n");
	while (getchar() != 10) {
		reg.value = 8;  /* rising */
		if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
			close(fd);
			sigaction(SIGALRM, &act_old, NULL);
			puts("Failure of rising.");
			return FAILURE;
		}
		putchar('-');
		reg.value = 0;  /* falling */
		if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
			close(fd);
			sigaction(SIGALRM, &act_old, NULL);
			puts("Failure of falling.");
			return FAILURE;
		}
		putchar('_');
	}

	usleep(100);

	printf("\nGet interrupt signal for %d times.\n", sig_counter);

	close(fd);
	sigaction(SIGALRM, &act_old, NULL);
	puts("\nEnd of program.");
	return SUCCESS;
}
